# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from . import api, scale
from .util import All as __all__

__all__ = __all__(globals())
__all__.update([name, getattr(api, name)] for name in ['D', 'E', 'topitch', 'unit', 'V'])
__all__.update([name, getattr(scale, name)] for name in ['harmonicminor', 'major', 'naturalminor', 'octatonic', 'wholetone'])
del api, scale



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager
import threading, logging

log = logging.getLogger(__name__)
local = threading.local()
deleted = object()

class Config: pass

@contextmanager
def threadlocals(**kwargs):
    old = {name: getattr(local, name) if hasattr(local, name) else deleted for name in kwargs}
    for name, value in kwargs.items():
        setattr(local, name, value)
    try:
        yield
    finally:
        for name, value in old.items():
            if value is deleted:
                delattr(local, name)
            else:
                setattr(local, name, value)

@contextmanager
def catch(obj, *logargs):
    if not hasattr(obj, '_onfire'):
        obj._onfire = False
    try:
        yield
        obj._onfire = False
    except Exception:
        if not obj._onfire: # TODO: Show error if it has changed.
            log.exception(*logargs)
            obj._onfire = True

class All(list):

    def __init__(self, namespace):
        self.namespace = namespace

    def update(self, entries):
        for name, obj in entries:
            self.namespace[name] = obj
            self.append(name)

class Lazy: # XXX: Is this really the most maintainable way?

    def __init__(self, globalsdict, name):
        self._resolve = lambda: globalsdict[name]

    def __str__(self):
        return str(self._resolve())

    def __getattr__(self, name):
        return getattr(self._resolve(), name)

    def __call__(self, *args, **kwargs):
        return self._resolve()(*args, **kwargs)

    def __getitem__(self, key):
        return self._resolve()[key]

    def __iter__(self):
        return iter(self._resolve())

    def __add__(self, other):
        return self._resolve() + other

    def __sub__(self, other):
        return self._resolve() - other

    def __mul__(self, other):
        return self._resolve() * other

    def __lshift__(self, other):
        return self._resolve() << other

    def __and__(self, other):
        return self._resolve() & other

    def __or__(self, other):
        return self._resolve() | other



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from .parse import concat, EParse, Program, rebase, Script, StepScript, vector, VParse
from .util import local

def V(*script, step = 0, continuous = False):
    return concat(lambda *args: StepScript(*args, step), VParse(float, step, continuous), script, {})

def D(*script):
    return concat(Script, VParse(vector, 0, False), script, {})

def E(cls, *script, initargs = (), **kwargs):
    namespace = object()
    kwargs = {(namespace, name): value for name, value in kwargs.items()}
    return concat(Script, EParse(Program(cls, initargs), namespace), script, kwargs)

unit = E(None, 'z')

def topitch(degree):
    c = local.context
    return _topitch(c.get('scale'), c.get('mode'), c.get('tonic'), degree)

def _topitch(scale, mode, tonic, degree):
    mode = rebase(mode)
    return tonic - scale[mode] + float((scale << mode)[degree[0] * scale.len + degree[1]] + degree[2])



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from .api import E
from .transform import Interpreter
from .util import Config, Lazy
from .xtra import XTRA
from diapyr import types
import bisect, logging, threading

log = logging.getLogger(__name__)

class Context:

    lazyname = '_lazy' # XXX: More reliably avoid collision?
    deleted = object()

    @types(Config)
    def __init__(self, config, sections = [(E(XTRA, '11r1'),)]):
        self.fastglobals = self.slowglobals = dict(
            {self.lazyname: Lazy},
            tuning = config.tuning,
            mode = 1,
            speed = 16, # XXX: Needed when sections is empty?
            sections = sections,
        )
        self.snapshot = self.fastglobals.copy()
        self.fastupdates = self.slowupdates = {}
        self.cache = {}
        self.slowlock = threading.Lock()
        self.fastlock = threading.Lock()
        i = Interpreter(self.lazyname, self.slowglobals)
        self.interpret = i.interpret if config.Lurlene.lazy else i.justexec

    def update(self, text):
        addupdate = []
        delete = []
        with self.slowlock:
            with self.fastlock:
                self.fastglobals = self.slowglobals.copy()
                self.fastupdates = self.slowupdates.copy()
            before = self.slowglobals.copy()
            self.interpret(text) # XXX: Impact of modifying mutable objects?
            for name, value in self.slowglobals.items():
                if not (name in before and value is before[name]):
                    self.slowupdates[name] = value
                    addupdate.append(name)
            for name in before:
                if name not in self.slowglobals:
                    self.slowupdates[name] = self.deleted
                    delete.append(name)
            with self.fastlock:
                self.fastglobals = self.slowglobals
                self.fastupdates = self.slowupdates
        if addupdate:
            log.info("Add/update: %s", ', '.join(addupdate))
        if delete:
            log.info("Delete: %s", ', '.join(delete))
        if not (addupdate or delete):
            log.info('No change.')

    def flip(self):
        if self.slowlock.acquire(False):
            try:
                with self.fastlock:
                    self.snapshot = self.fastglobals.copy()
                    self.fastupdates.clear()
            finally:
                self.slowlock.release()

    class NoSuchGlobalException(Exception): pass

    def get(self, name):
        'Include changes made via global keyword.'
        with self.fastlock:
            # If the fastglobals value (or deleted) is due to update, return snapshot value (or deleted):
            value = self.fastglobals.get(name, self.deleted)
            if name in self.fastupdates and value is self.fastupdates[name]:
                try:
                    return self.snapshot[name]
                except KeyError:
                    raise self.NoSuchGlobalException(name) # TODO: Test this case.
            if value is self.deleted:
                raise self.NoSuchGlobalException(name)
            return value

    def _cachedproperty(f):
        name = f.__name__
        code = f.__code__
        params = code.co_varnames[1:code.co_argcount]
        def fget(self):
            args = [self.get(p) for p in params]
            try:
                cacheargs, value = self.cache[name]
                if all(x is y for x, y in zip(cacheargs, args)):
                    return value
            except KeyError:
                pass
            value = f(*[self] + args)
            self.cache[name] = args, value
            return value
        return property(fget)

    @_cachedproperty
    def sections(self, speed, sections):
        return Sections(speed, sections)

class Sections:

    def __init__(self, speed, sections):
        def sectionends():
            end = 0
            for section in sections:
                end += speed * max(pattern.len for pattern in section)
                yield end
        # FIXME: Recalculate sectionends when a lazy section is resized.
        self.sectionends = list(sectionends())
        self.sections = sections

    @property
    def totalframecount(self):
        return self.sectionends[-1]

    def startframe(self, sectionindex):
        return self.sectionends[sectionindex - 1] if sectionindex else 0

    def sectionandframe(self, frameindex):
        localframe = frameindex % self.sectionends[-1]
        i = bisect.bisect(self.sectionends, localframe)
        return self.sections[i], localframe - self.startframe(i)



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from .context import Context
from .util import Config
from diapyr import types
from splut.bg import SimpleBackground
import logging, socket, timelyOSC

log = logging.getLogger(__name__)

class OSCClient:

    def __init__(self, host, port, bufsize, handlers):
        self.address = host, port
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # XXX: Close it?
        self.sock.bind(self.address)
        self.bufsize = bufsize
        self.handlers = handlers

    def pumponeortimeout(self):
        try:
            bytes, address = self.sock.recvfrom(self.bufsize)
            self._message(address, [], timelyOSC.parse(bytes))
        except socket.timeout:
            pass
        except Exception:
            log.exception('Failed to receive message:')

    def _message(self, udpaddr, timetags, message):
        try:
            addrpattern = message.addrpattern
        except AttributeError:
            self._elements(udpaddr, timetags + [message.timetag], message.elements)
            return
        try:
            handler = self.handlers[addrpattern]
        except KeyError:
            log.warning("Unhandled message: %s", message)
            return
        handler.handle(timetags, message, lambda reply: self.sock.sendto(reply, udpaddr))

    def _elements(self, udpaddr, timetags, elements):
        for element in elements:
            self._message(udpaddr, timetags, element)

    def interrupt(self):
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.sendto(timelyOSC.Message('/interrupt', []).ser(), self.address)

class Handler: pass

class OSCListen(SimpleBackground):

    @types(Config, [Handler])
    def __init__(self, config, handlers):
        super().__init__(config.profile)
        self.config = config
        self.handlers = {a: h for h in handlers for a in h.addresses}

    def start(self):
        config = self.config.OSC
        super().start(self.bg, OSCClient(
                *(getattr(config, name) for name in ['host', 'port', 'bufsize']),
                self.handlers))

    def bg(self, client):
        while not self.quit:
            client.pumponeortimeout()

class LurleneHandler(Handler):

    addresses = '/lurlene',

    @types(Context)
    def __init__(self, context):
        self.context = context

    def handle(self, timetags, message, reply):
        try:
            text, = message.args
            self.context.update(text)
        except Exception:
            log.exception('Update failed:')

class InterruptHandler(Handler):

    addresses = '/interrupt',

    @types()
    def __init__(self):
        pass

    def handle(self, timetags, message, reply):
        pass # Do nothing.

def configure(di):
    di.add(LurleneHandler)
    di.add(InterruptHandler)
    di.add(OSCListen)



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from .context import Sections, Context
from .util import catch, Config, threadlocals
from diapyr import types
from foyndation import innerclass, singleton
from functools import partial
from itertools import zip_longest
import logging, bisect, difflib

log = logging.getLogger(__name__)

class NoSuchSectionException(Exception): pass

class Channel:

    def __init__(self, index, nametoproxy):
        self.letter = chr(ord('A') + index)
        self.nametoproxy = nametoproxy

class LiveCodingBridge:

    bias = .5 # TODO: Make configurable for predictable swing in odd speed case.

    @types(Config, Context)
    def __init__(self, config, context):
        self.loop = not config.ignoreloop
        self.sectionname = config.section
        self.context = context

    @property
    def pianorollheight(self):
        return self.context.get('speed')

    @innerclass
    class Session:

        def __init__(self, chips):
            self.channels = [Channel(index, {name: proxy for name, proxy in zip(chips, proxies) if proxy is not None})
                    for index, proxies in enumerate(zip_longest(*chips.values()))]

        def _quiet(self):
            for channel in self.channels:
                for proxy in channel.nametoproxy.values():
                    proxy.blank()

        def _step(self, speed, section, frame):
            self._quiet()
            for channel, pattern in zip(self.channels, section):
                with catch(channel, "Channel %s update failed:", channel.letter):
                    pattern.apply(speed, frame, channel.nametoproxy)

    def _initialframe(self):
        if self.sectionname is None:
            sectionindex = 0
        else:
            section = getattr(self.context, self.sectionname)
            try:
                sectionindex = self.context.get('sections').index(section)
            except ValueError:
                raise NoSuchSectionException(self.sectionname)
        return self.context.sections.startframe(sectionindex)

    def frames(self, chips):
        session = self.Session(chips)
        frameindex = self._initialframe() + self.bias
        with threadlocals(context = self.context):
            while self.loop or frameindex < self.context.sections.totalframecount:
                oldspeed = self.context.get('speed')
                oldsections = self.context.get('sections')
                frame = session._quiet
                if self.context.sections.totalframecount: # Otherwise freeze until there is something to play.
                    with catch(session, 'Failed to prepare a frame:'):
                        frame = partial(session._step, self.context.get('speed'), *self.context.sections.sectionandframe(frameindex))
                        frameindex += 1
                frame()
                yield
                self.context.flip()
                if oldspeed != self.context.get('speed'):
                    frameindex = (frameindex - self.bias) / oldspeed * self.context.get('speed') + self.bias
                if oldsections != self.context.get('sections'):
                    frameindex = self._adjustframeindex(Sections(self.context.get('speed'), oldsections), frameindex)

    def _adjustframeindex(self, oldsections, frameindex):
        baseframe = (frameindex // oldsections.totalframecount) * self.context.sections.totalframecount
        localframe = frameindex % oldsections.totalframecount
        oldsectionindex = bisect.bisect(oldsections.sectionends, localframe)
        sectionframe = localframe - oldsections.startframe(oldsectionindex)
        opcodes = difflib.SequenceMatcher(a = oldsections.sections, b = self.context.get('sections')).get_opcodes()
        @singleton
        def sectionindexandframe():
            for tag, i1, i2, j1, j2 in opcodes:
                if 'equal' == tag and i1 <= oldsectionindex and oldsectionindex < i2:
                    return j1 + oldsectionindex - i1, sectionframe
            oldsection = oldsections.sections[oldsectionindex]
            for tag, i1, i2, j1, j2 in opcodes:
                if 'insert' == tag and oldsection in self.context.get('sections')[j1:j2]:
                    return j1 + self.context.get('sections')[j1:j2].index(oldsection), sectionframe
            for tag, i1, i2, j1, j2 in opcodes:
                if tag in {'delete', 'replace'} and i1 <= oldsectionindex and oldsectionindex < i2:
                    return j1, self.bias
            return 0, self.bias # TODO: Test this case
        return baseframe + self.context.sections.startframe(sectionindexandframe[0]) + sectionindexandframe[1]



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from .api import V

harmonicminor = V('0 2 3 5 7 8 11', step = 12, continuous = True)
major = V('0 2 4 5 7 9 11', step = 12, continuous = True)
naturalminor = V('0 2 3 5 7 8 10', step = 12, continuous = True)
octatonic = V('0 1 3 4 6 7 9 10', step = 12, continuous = True)
wholetone = V('0 2 4 6 8 10', step = 12, continuous = True)



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from .api import D, _topitch, V
from .pitch import E4
from .scale import major

class XTRA:

    degree = D('-') + D('- 5- 1 5,+').of(6)
    envflag = V('30x,1')
    mute = False

    def on(self, ym, frame):
        if self.mute:
            return
        envflag = self.envflag[frame]
        pitch = _topitch(major, 1, E4, self.degree[frame])
        for chan in range(min(3, len(ym._chanproxies))):
            ym[chan].toneflag = True
            ym[chan].level = 15
            ym[chan].envflag = envflag
            ym[chan].tonepitch = pitch
            ym[chan].toneperiod += chan * 2
        if envflag and not self.envflag[frame - 1]:
            ym.envshape = 0
        ym.envperiod = 30 << 8

    def off(self):
        type(self).mute = True



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from math import floor
import bisect, numpy as np

class BaseSegment:

    def __init__(self, initial, excess, width):
        self.initial = initial
        self.excess = excess
        self.width = width

    def getvalue(self, frame, xadjust):
        return self.unbiased(frame)

class FlatSegment(BaseSegment):

    def unbiased(self, frame):
        return self.initial

    def settarget(self, target):
        pass

class Segment(BaseSegment):

    def unbiased(self, frame):
        return self.initial + (self.excess + frame) * self.perframe

    def settarget(self, target):
        self.perframe = (target - self.initial) / (self.excess + self.width)

class BiasSegment(Segment):

    def getvalue(self, frame, xadjust):
        return super().getvalue(frame, xadjust) + self.bias

    def settarget(self, target):
        super().settarget(target)
        self.bias = float(np.sign(self.perframe)) / -2

class Segments:

    def __init__(self):
        self.frames = [] # First is always 0, but it's convenient.
        self.segments = []
        self.len = 0

    def add(self, segment):
        self.frames.append(self.len)
        self.segments.append(segment)
        self.len += segment.width

    def init(self, segments, initials):
        self.frames = segments.frames.copy()
        self.segments = [type(s)(i, s.excess, s.width) for s, i in zip(segments.segments, initials)]
        self.len = segments.len
        n = len(self.segments)
        for i in range(n):
            if i + 1 < n:
                t = self.segments[i + 1]
            else:
                t = self.segments[0]
            self.segments[i].settarget(t.initial)

    def empty(self):
        return not self.frames

    def at(self, index):
        return self.frames[index], self.segments[index]

    def forframe(self, absframe, xadjust):
        relframe = absframe % self.len
        sectionframe, section = self.at(bisect.bisect(self.frames, relframe) - 1)
        return section.getvalue(relframe - sectionframe, absframe - relframe + xadjust)

class Operators:

    def __getitem__(self, frame):
        return Slice(self, frame) if isinstance(frame, slice) else self.getitem(frame, 0)

    def __add__(self, that):
        return Sum(self, that)

    def __sub__(self, that):
        return Diff(self, that)

    def __mul__(self, that):
        return self.mulcls(self, that)

    def __and__(self, that):
        # FIXME: Suspect loss of laziness of self is causing namespace-related kwarg-passing problems.
        return Merge(self, that)

    def __or__(self, that):
        return Then(self, that, self, that)

    def __lshift__(self, frames):
        return self >> -frames

    def __rshift__(self, frames):
        return RShift(self, frames)

    def of(self, beat):
        return Of(self, beat)

    def inversions(self):
        degreetoindex = {}
        degrees = []
        refs = []
        for s in self.segments.segments:
            degree = tuple(s.initial[1:])
            if degree in degreetoindex:
                index = degreetoindex[degree]
            else:
                degreetoindex[degree] = index = len(degrees)
                degrees.append(np.array((0,) + degree))
            refs.append((index, s.initial[0]))
        invs = []
        for _ in range(len(degrees)):
            segments = Segments()
            segments.init(self.segments, (degrees[index] + np.array([octave, 0, 0]) for index, octave in refs))
            invs.append(type(self)(segments, self.kwargs))
            degrees.append(degrees.pop(0) + np.array([1, 0, 0]))
        return invs

    def apply(self, speed, frame, chips):
        self.of(speed)[frame].perform(frame, speed, chips, self.kwargs)

class EventSegment:

    def __init__(self, relframe, onframes, width, program, namespace):
        self.relframe = relframe
        self.onframes = onframes
        self.width = width
        self.program = program
        self.namespace = namespace

    def getvalue(self, frame, xadjust):
        return Event(xadjust + self.relframe, self.onframes, self.program, self.namespace)

class Event:

    def __init__(self, absframe, onframes, program, namespace):
        self.absframe = absframe
        self.onframes = onframes
        self.program = program
        self.namespace = namespace

    def perform(self, frame, speed, chips, kwargs):
        note = self.program.new() # XXX: Allow a note to maintain state?
        def noteargs(params, shift, **extras):
            for name in params:
                if name in extras:
                    yield name, extras[name]
                elif 'frame' == name:
                    yield name, Frame(frame - self.absframe * speed + shift)
                else:
                    key = self.namespace, name
                    if key in kwargs:
                        yield name, (kwargs[key] >> -self.absframe).of(speed) >> shift
        if self.onframes is None:
            if self.program.onparams is not None:
                note.on(**dict(noteargs(self.program.onparams, 0, **chips, onframes = None)))
        else:
            if self.program.offparams is not None:
                onframes = self.onframes * speed
                note.off(**dict(noteargs(self.program.offparams, onframes, **chips, onframes = onframes)))
            elif self.program.onparams is not None:
                onframes = self.onframes * speed
                note.on(**dict(noteargs(self.program.onparams, onframes, **chips, onframes = onframes)))

class Frame(float):

    def pick(self, v):
        return v[floor(self)]

    def __mod__(self, other):
        return type(self)(super().__mod__(other))

class Overlay:

    def __init__(self, e1, e2):
        self.e1 = e1
        self.e2 = e2

    def perform(self, *args):
        for e in self.e1, self.e2:
            e.perform(*args)

class Binary(Operators):

    def __init__(self, p1, p2):
        self.p1 = p1
        self.p2 = p2

class Sum(Binary):

    kwargs = {} # XXX: Can this ever have kwargs? Or replace with bundle mechanism.

    @property
    def len(self):
        return max(self.p1.len, self.p2.len)

    def getitem(self, frame, shift):
        return self.p1.getitem(frame, shift) + self.p2.getitem(frame, shift)

class Diff(Binary):

    @property
    def len(self):
        return max(self.p1.len, self.p2.len)

    def getitem(self, frame, shift):
        return self.p1.getitem(frame, shift) - self.p2.getitem(frame, shift)

class Mul(Binary):

    @property
    def len(self):
        return max(self.p1.len, self.p2.len)

    def getitem(self, frame, shift):
        return self.p1.getitem(frame, shift) * self.p2.getitem(frame, shift)

class Merge(Binary):

    @property
    def mulcls(self):
        mulcls, = set(p.mulcls for p in [self.p1, self.p2])
        return mulcls

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.len = max(self.p1.len, self.p2.len)
        self.kwargs = {**self.p1.kwargs, **self.p2.kwargs}

    def getitem(self, frame, shift):
        x = self.p1.getitem(frame, shift)
        y = self.p2.getitem(frame, shift)
        return Overlay(x, y)

class RShift(Operators):

    @property
    def mulcls(self):
        return self.p.mulcls

    @property
    def kwargs(self):
        return self.p.kwargs

    @property
    def len(self):
        return self.p.len

    def __init__(self, p, frames):
        self.p = p
        self.frames = frames

    def getitem(self, frame, shift):
        return self.p.getitem(frame, self.frames + shift)

class Repeat(Operators):

    @property
    def mulcls(self):
        return self.p.mulcls

    @property
    def kwargs(self):
        return self.p.kwargs

    def __init__(self, p, times):
        self.len = p.len * times
        self.p = p
        self.times = times

    def getitem(self, frame, shift):
        return self.p.getitem(frame, shift)

class Of(Operators):

    kwargs = {} # TODO: Implement.

    @property
    def len(self):
        return self.p.len * self.beat

    @property
    def mulcls(self):
        return self.p.mulcls

    def __init__(self, p, beat):
        self.p = p
        self.beat = beat

    def getitem(self, frame, shift):
        return self.p.getitem(frame / self.beat, shift / self.beat)

class Concat(Binary):

    kwargs = {} # TODO LATER: Implement.

    @property
    def len(self):
        return self.p1.len + self.p2.len

    @property
    def mulcls(self):
        mulcls, = set(p.mulcls for p in [self.p1, self.p2])
        return mulcls

    def getitem(self, frame, shift):
        split = self.p1.segments.len
        if frame - shift < split:
            return self.p1.getitem(frame, shift)
        else:
            return self.p2.getitem(frame, shift + split)

class Then(Binary):

    @property
    def mulcls(self):
        mulcls, = set(p.mulcls for p in [self.p1, self.p2])
        return mulcls

    @property
    def len(self):
        return self.len1.len + self.len2.len

    def __init__(self, p1, p2, len1, len2):
        super().__init__(p1, p2)
        self.kwargs = {name: Then(
            p1.kwargs.get(name, p2.kwargs.get(name)),
            p2.kwargs.get(name, p1.kwargs.get(name)),
            len1,
            len2,
        ) for name in p1.kwargs.keys() | p2.kwargs.keys()}
        self.len1 = len1
        self.len2 = len2

    def getitem(self, frame, shift):
        split = self.len1.len
        total = split + self.len2.len
        loop = (frame - shift) // total
        if (frame - shift) % total < split:
            return self.p1.getitem(frame, shift + self.len2.len * loop)
        else:
            return self.p2.getitem(frame, shift + self.len1.len * (1 + loop))

class Slice(Operators):

    @property
    def mulcls(self):
        return self.p.mulcls

    def __init__(self, p, slice):
        def readslice(ref, adj):
            return ref if adj is None else (ref + adj if adj < 0 else adj)
        self.start = readslice(0, slice.start)
        self.stop = readslice(p.len, slice.stop)
        # TODO: Use step.
        self.len = self.stop - self.start
        self.kwargs = {name: Slice(v, slice) for name, v in p.kwargs.items()}
        self.p = p

    def getitem(self, frame, shift):
        loop = (frame - shift) // self.len
        return self.p.getitem(frame, shift - self.start - (self.p.len - self.len) * loop)



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from .model import Operators, Segments, FlatSegment, BiasSegment, Segment, Concat, EventSegment, Repeat, Mul
from foyndation import innerclass
from fractions import Fraction
import re, numpy as np, inspect, itertools

class Script(Operators):

    mulcls = Repeat

    def __init__(self, segments, kwargs):
        self.len = max(itertools.chain([segments.len], (s.len for s in kwargs.values())))
        self.segments = segments
        self.kwargs = kwargs

    def getitem(self, frame, shift):
        return self.segments.forframe(frame - shift, shift)

class StepScript(Script):

    mulcls = Mul

    def __init__(self, segments, kwargs, step):
        super().__init__(segments, kwargs)
        self.step = step

    def getitem(self, frame, shift):
        return Value(super().getitem(frame, shift) + (frame - shift) // self.segments.len * self.step)

class Value(float):

    def pick(self, sequence):
        return sequence[round(self)]

class BadWordException(Exception): pass

class Parse:

    def parse(self, script, successor):
        session = self.Session()
        session.segments = Segments()
        for word in re.findall(r'[^\s|]+', script):
            session.parseword(word)
        session.wrap(successor)
        return session.segments

class VParse(Parse):

    pattern = re.compile('(?:([0-9.]+)x)?(-?[0-9.]+)?(#+|b+)?([+]+|-+)?(?:/(/)?([0-9.]*))?')

    def __init__(self, type, step, continuous):
        self.type = type
        self.step = step
        self.continuous = continuous

    @innerclass
    class Session:

        def parseword(self, word):
            m = self.pattern.fullmatch(word)
            if m is None:
                raise BadWordException(word)
            width = m.group(1)
            width = 1 if width is None else float(width)
            initial = m.group(2)
            initial = self.type() if initial is None else self.type(initial)
            acc = m.group(3)
            if acc is not None:
                initial[2] += (1 if '#' in acc else -1) * len(acc)
            octave = m.group(4)
            if octave is not None:
                initial[0] += (1 if '+' in octave else -1) * len(octave)
            bias = m.group(5)
            slide = m.group(6)
            slide = (width if self.continuous else 0) if slide is None else (float(slide) if slide else width)
            if not self.segments.empty():
                self._wrap(initial)
            hold = width - slide
            if hold > 0 or (not hold and not slide):
                self.segments.add(FlatSegment(initial, None, hold))
            if slide:
                self.segments.add((BiasSegment if bias else Segment)(initial, max(0, slide - width), min(width, slide)))

        def wrap(self, successor):
            if successor is None:
                _, firstsection = self.segments.at(0)
                self._wrap(firstsection.initial + self.step)
            else:
                self._wrap(successor[0]) # TODO: Unit-test what effect step would have.

        def _wrap(self, value):
            _, lastsection = self.segments.at(-1)
            lastsection.settarget(value)

def rebase(n, frombase = 1):
    return np.sign(n) * max(0, abs(n) - frombase)

def vector(dstr = None):
    return np.array([0, 0 if dstr is None else rebase(float(dstr)), 0])

def _flatten(scriptforest):
    def g(scriptforest):
        for textorseq in scriptforest:
            if hasattr(textorseq, 'encode'):
                yield textorseq
            else:
                yield from g(textorseq)
    return ' '.join(map(str, g(scriptforest)))

def concat(scriptcls, parser, scriptforest, kwargs):
    scripts = []
    successor = None
    for act in reversed(_flatten(scriptforest).split(',')):
        successor = scriptcls(parser.parse(act, successor), kwargs)
        scripts.insert(0, successor)
    return scripts[0] if 1 == len(scripts) else Concat(*scripts)

def _readnumber(numstr, default):
    if numstr:
        if '-' == numstr[0]:
            sign = -1
            numstr = numstr[1:]
        else:
            sign = 1
        return sign * Fraction(f"1{numstr}" if '/' == numstr[0] else numstr)
    return default

class EParse(Parse):

    pattern = re.compile('(?:([0-9]+)x)?(-?[0-9./]+)?(?:r([0-9./]*))?|([0-9./]*)z')

    def __init__(self, program, namespace):
        self.program = program
        self.namespace = namespace

    @innerclass
    class Session:

        def parseword(self, word):
            m = self.pattern.fullmatch(word)
            if m is None:
                raise BadWordException(word)
            rest = m.group(4)
            if rest is not None:
                self.segments.add(EventSegment(self.segments.len, None, _readnumber(rest, 1), silence, self.namespace))
                return
            count = m.group(1)
            count = 1 if count is None else int(count)
            if count < 1:
                raise BadWordException(word)
            width = _readnumber(m.group(2), 1)
            offwidth = m.group(3)
            if offwidth is None:
                offwidth = 0
            elif offwidth:
                offwidth = min(width, _readnumber(offwidth, None))
            else:
                offwidth = width
            onwidth = width - offwidth
            for _ in range(count):
                if onwidth:
                    self.segments.add(EventSegment(self.segments.len, None, onwidth, self.program, self.namespace))
                if offwidth:
                    self.segments.add(EventSegment(self.segments.len, onwidth, offwidth, self.program, self.namespace))

        def wrap(self, successor):
            pass

class Program:

    def __init__(self, cls, init):
        # FIXME: If cls is Lazy these params will get out of date.
        def params(name):
            try:
                unbound = getattr(cls, name)
            except AttributeError:
                return
            return list(itertools.islice(inspect.signature(unbound).parameters.keys(), 1, None))
        self.onparams = params('on')
        self.offparams = params('off')
        self.cls = cls
        self.init = init

    def new(self):
        return self.cls(*self.init)

silence = Program(object, ())



from setuptools import setup
try:
    from Cython.Build import cythonize
except ModuleNotFoundError:
    pass
from pathlib import Path
from setuptools import find_packages
import os

class SourceInfo:

    class PYXPath:

        def __init__(self, module, path):
            self.module = module
            self.path = path

        def make_ext(self):
            g = {}
            with open(f"{self.path}bld") as f: # Assume project root.
                exec(f.read(), g)
            return g['make_ext'](self.module, self.path)

    def __init__(self, rootdir):
        def addextpaths(dirpath, moduleprefix, suffix = '.pyx'):
            for name in sorted(os.listdir(os.path.join(rootdir, dirpath))):
                if name.endswith(suffix):
                    module = f"{moduleprefix}{name[:-len(suffix)]}"
                    if module not in extpaths:
                        extpaths[module] = self.PYXPath(module, os.path.join(dirpath, name))
        self.packages = find_packages(rootdir)
        extpaths = {}
        addextpaths('.', '')
        for package in self.packages:
            addextpaths(package.replace('.', os.sep), f"{package}.")
        self.extpaths = extpaths.values()

    def setup_kwargs(self):
        kwargs = dict(packages = self.packages)
        try:
            kwargs['long_description'] = Path('README.md').read_text()
            kwargs['long_description_content_type'] = 'text/markdown'
        except FileNotFoundError:
            pass
        if self.extpaths:
            kwargs['ext_modules'] = cythonize([path.make_ext() for path in self.extpaths])
        return kwargs

sourceinfo = SourceInfo('.')
setup(
    name = 'Lurlene',
    version = '21',
    description = 'Python-based live-coding language optimised for a small number of channels',
    url = 'https://github.com/combatopera/Lurlene',
    author = 'Andrzej Cichocki',
    author_email = '3613868+combatopera@users.noreply.github.com',
    py_modules = [],
    install_requires = ['diapyr>=18', 'foyndation>=12', 'numpy>=2.0.2', 'splut>=4', 'timelyOSC>=4'],
    package_data = {'': ['*.pxd', '*.pyx', '*.pyxbld', '*.arid', '*.aridt']},
    entry_points = {'console_scripts': []},
    **sourceinfo.setup_kwargs(),
)



# Copyright 2019 Andrzej Cichocki

# This file is part of Lurlene.
#
# Lurlene is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lurlene is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lurlene.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
from foyndation import innerclass
import ast, logging

log = logging.getLogger(__name__)

class Interpreter:

    @innerclass
    class Transform(ast.NodeTransformer):

        def __init__(self):
            self.lazycounts = defaultdict(lambda: 0)

        def visit_ClassDef(self, node):
            node.body[:] = (self.visit(statement) for statement in node.body)
            return node

        def visit_Name(self, node):
            if not isinstance(node.ctx, ast.Load):
                return node
            name = node.id
            if name not in self.globalsdict:
                return node
            self.lazycounts[name] += 1
            return ast.Call(ast.Name(self.lazyname, ast.Load()), [ast.Call(ast.Name('globals', ast.Load()), [], []), ast.Str(name)], [])

        def report(self):
            if self.lazycounts:
                log.debug("Lazy: %s", ', '.join(f"""{n}{f"*{c}" if 1 != c else ''}""" for n, c in self.lazycounts.items()))

    def __init__(self, lazyname, globalsdict):
        self.lazyname = lazyname
        self.globalsdict = globalsdict

    def interpret(self, text):
        transform = self.Transform()
        for statement in ast.parse(text).body:
            self.justexec(ast.fix_missing_locations(ast.Module([transform.visit(statement)], []))) # XXX: Are locations accurate?
        transform.report()

    def justexec(self, textorast):
        exec(compile(textorast, '<text>', 'exec'), self.globalsdict)
